import csv
import datetime
import json
import logging
import traceback

import dateutil.parser
import requests
import urllib3
import hashlib
import pytiqs.exceptions as ex

import pytiqs.constants as constants
from pytiqs.constants import Variety, Exchange, ProductType, TransactionType, OrderType, Retention

from typing import Dict, Any, Optional, List, Union
from six import StringIO, PY2
from pytiqs.__version__ import __version__, __title__

log = logging.getLogger(__name__)


class Tiqs(object):
    _default_timeout = 5  # type: int
    _token = None

    def __init__(
            self, app_id: str,
            timeout: Optional[int] = None,
            debug: bool = False,
            root_url: Optional[str] = None,
            pool: Optional[Dict[str, Any]] = None
    ) -> None:
        self.app_id = app_id
        self.debug = debug
        self.disable_ssl = False
        self.timeout = timeout or self._default_timeout
        self.proxies = {}  # type: Dict[str, str]
        self._routes = constants.Routes(root_url)

        # Create requests session by default
        self.req_session = requests.Session()  # type: requests.Session
        if pool:
            req_adapter = requests.adapters.HTTPAdapter(**pool)
            self.req_session.mount("https://", req_adapter)

        # disable requests SSL warning
        urllib3.disable_warnings()

    def login_url(self) -> str:
        return "%s?appId=%s" % (self._routes.get_home_url(), self.app_id)

    def generate_session(self, request_token: str, api_secret: str) -> Dict[str, str]:
        checksum_formula = "{app_id}:{api_secret}:{request_token}".format(app_id=self.app_id, api_secret=api_secret,
                                                                          request_token=request_token)
        checksum = hashlib.sha256(checksum_formula.encode("utf-8")).hexdigest()

        resp = self._post(self._routes.AUTHENTICATE_TOKEN, params={
            "appId": self.app_id,
            "token": request_token,
            "checkSum": checksum
        })  # type: Dict[str, str]

        if "token" in resp:
            if self.debug:
                log.debug("token received from generate_session: " + resp["token"])
            self.set_token(resp["token"])
        return resp

    def set_token(self, value: str) -> None:
        self._token = value

    def get_token(self) -> str:
        return self._token or ""

    def invalidate_session(self) -> None:
        self._token = None

    def get_instruments(self) -> Any:
        return self._parse_instrument(self._get(self._routes.ALL_INSTRUMENTS))

    def user_details(self) -> Dict[str, Any]:
        return self._get(self._routes.USER_DETAILS)

    def user_limits(self) -> List[Dict[str, Any]]:
        return self._parse_user_limits(self._get(self._routes.GET_USER_LIMITS))

    def holidays(self) -> Dict[str, Any]:
        return self._get(self._routes.HOLIDAYS_LIST)

    def index_list(self) -> List[Dict[str, str]]:
        return self._get(self._routes.INDEX_LIST)

    def historical_data(self,
                        exchange: str,
                        instrument_token: str,
                        from_time: datetime.datetime,
                        to_time: datetime.datetime,
                        interval: str,
                        ) -> List[Any]:
        from_string = from_time.strftime('%Y-%m-%dT%H:%M:%S')
        to_string = to_time.strftime('%Y-%m-%dT%H:%M:%S')
        data = self._get(
            self._routes.GET_HISTORICAL_DATA,
            url_args={"exchange": exchange, "token": instrument_token, "interval": interval},
            params={"from": from_string, "to": to_string}
        )
        return self._parse_historical_data(data)

    def option_chain(self, params: Dict[str, str]) -> List[Dict[str, Any]]:
        return self._parse_option_chain(self._post(self._routes.OPTION_CHAIN, params=params))

    def option_chain_symbols(self):
        return self._get(self._routes.OPTION_CHAIN_SYMBOLS)

    def greeks(self, instrument_tokens: List[int]):
        return self._parse_greeks(instrument_tokens, self._post(self._routes.GREEKS, params=instrument_tokens))

    def order_margin(self, params: Dict[str, str]) -> Dict[str, Any]:
        return self._post(self._routes.ORDER_MARGIN, params=params)

    def basket_order_margin(self, params: List[Dict[str, str]]) -> Dict[str, str]:
        return self._post(self._routes.BASKET_ORDER_MARGIN, params=params)

    def get_order(self, order_id: str) -> List[Dict[str, Any]]:
        return self._format_order_by_id_response(
            self._get(self._routes.GET_ORDER_BY_ID, url_args={"order_id": order_id}))

    def get_user_orders(self) -> List[Dict[str, Any]]:
        return self._format_user_orders_response(self._get(self._routes.GET_USER_ORDERS))

    def get_user_trades(self) -> List[Dict[str, Any]]:
        return self._format_user_trades_response(self._get(self._routes.GET_USER_TRADES))

    def get_positions(self) -> List[Dict[str, Any]]:
        return self._get(self._routes.GET_USER_POSITIONS)

    def place_order(
            self,
            exchange: Exchange,
            token: str,
            qty: int,
            disclosed_qty: int,
            product: ProductType,
            symbol: str,
            transaction_type: TransactionType,
            order_type: OrderType,
            variety: Variety,
            price: float,
            validity: Retention,
            tags: Optional[str] = None,
            amo: bool = False,
            trigger_price: Optional[float] = None
    ) -> str:
        params = dict(
            exchange=str(exchange),
            token=token,
            quantity="{qty}".format(qty=qty),
            disclosedQty="{disclosed_qty}".format(disclosed_qty=disclosed_qty),
            product=str(product),
            symbol=symbol,
            transactionType=str(transaction_type),
            order=str(order_type),
            price="{price}".format(price=price),
            validity=str(validity),
            tags=tags or "",
            amo=amo,
            triggerPrice="{trigger_price}".format(trigger_price=trigger_price or "")
        )  # type: Dict[str, Any]
        return self._post(
            self._routes.PLACE_ORDER,
            url_args={"variety": str(variety)},
            params=params
        )["orderNo"]

    def modify_order_by_id(
            self,
            order_id: str,
            exchange: Exchange,
            token: str,
            qty: int,
            disclosed_qty: int,
            product: ProductType,
            transaction_type: TransactionType,
            order_type: OrderType,
            price: float,
            validity: Retention,
            amo: bool = False,
            trigger_price: Optional[float] = None,
            tags: Optional[str] = None,
            book_loss_price: Optional[float] = None,
            book_profit_price: Optional[float] = None
    ) -> str:
        params = dict(
            exchange=str(exchange),
            token=token,
            quantity="{qty}".format(qty=qty),
            disclosedQty="{disclosed_qty}".format(disclosed_qty=disclosed_qty),
            product=str(product),
            transactionType=str(transaction_type),
            order=str(order_type),
            price="{price}".format(price=price),
            validity=str(validity),
            tags=tags or "",
            amo=amo,
            triggerPrice="{trigger_price}".format(trigger_price=trigger_price or "0"),
            bookLossPrice="{book_loss_price}".format(book_loss_price=book_loss_price or ""),
            bookProfitPrice="{book_profit_price}".format(book_profit_price=book_profit_price or "")
        )  # type: Dict[str, Any]
        return self._patch(
            self._routes.UPDATE_ORDER_BY_ID,
            url_args={"order_id": order_id},
            params=params
        )["message"]

    def single_instrument_quote(self, mode: str, instrument_token: int):
        data = dict(token=instrument_token)
        return self._post(self._routes.QUOTE_DATA, url_args=dict(mode=mode), params=data)

    def multiple_instrument_quotes(self, mode: str, instrument_tokens: List[int]):
        return self._post(self._routes.QUOTES_DATA, url_args=dict(mode=mode), params=instrument_tokens)

    def delete_order(self, order_id: str) -> str:
        return self._delete(self._routes.DELETE_ORDER, url_args={"order_id": order_id})["message"]

    def _parse_instrument(self, data: Any) -> Any:
        d = data
        if not PY2 and isinstance(d, bytes):
            d = data.decode("utf-8").strip()  # Decode unicode data

        records = []
        reader = csv.DictReader(StringIO(d))
        for row in reader:
            try:
                if row["LotSize"]:
                    row["LotSize"] = int(row["LotSize"])
                if row["TickSize"]:
                    row["TickSize"] = int(row["TickSize"])
                if row["PricePrecision"]:
                    row["PricePrecision"] = int(row["PricePrecision"])
                if row["Multiplier"]:
                    row["Multiplier"] = int(row["Multiplier"])
                if row["PriceMultiplier"]:
                    row["PriceMultiplier"] = float(row["PriceMultiplier"])
                if row["StrikePrice"]:
                    row["StrikePrice"] = float(row["StrikePrice"])
                records.append(row)
            except ValueError:
                if self.debug:
                    log.error(json.dumps(row), traceback.format_exc())
        return records

    def _get(self, route: str, url_args: Optional[Dict[str, str]] = None, params: Any = None) -> Any:
        return self._request(route, "GET", url_args=url_args, params=params)

    def _post(self,
              route: str,
              url_args: Optional[Dict[str, Any]] = None,
              params: Any = None,
              query_params: Any = None) -> Any:
        return self._request(route, "POST", url_args=url_args, params=params, query_params=query_params)

    def _patch(self,
               route: str,
               url_args: Optional[Dict[str, Any]] = None,
               params: Any = None,
               query_params: Any = None) -> Any:
        return self._request(route, "PATCH", url_args=url_args, params=params,
                             query_params=query_params)

    def _delete(self, route: str, url_args: Optional[Dict[str, str]] = None, params: Any = None,
                is_json: bool = False) -> Any:
        return self._request(route, "DELETE", url_args=url_args, params=params)

    def _request(self, route: str,
                 method: str,
                 url_args: Optional[Dict[str, str]] = None,
                 params: Any = None,
                 query_params: Any = None) -> Any:
        if url_args:
            uri = route.format(**url_args)
        else:
            uri = route

        url = uri

        headers = {
            "User-Agent": self._user_agent(),
            "appId": self.app_id
        }

        if self.token:
            headers["token"] = self.token

        if self.debug:
            log.debug(f"Request: {method} {url} {json.dumps(params)} {json.dumps(headers)}")

        if params is not None and method in ["GET", "DELETE"]:
            query_params = params

        try:
            r = self.req_session.request(method,
                                         url,
                                         json=params if (method in ["POST", "PUT", "PATCH"]) else None,
                                         data=None,
                                         params=query_params,
                                         headers=headers,
                                         verify=not self.disable_ssl,
                                         allow_redirects=True,
                                         timeout=self.timeout,
                                         proxies=self.proxies)
        except Exception as e:
            raise e

        if self.debug:
            log.debug('response: ' + method + " " + url + " " + f"\n{r.content[:1000]!r}")
        if "json" in r.headers["content-type"]:
            try:
                data = r.json()
            except ValueError:
                raise ex.DataException(f"Couldn't parse the JSON response received from the server: {r.content!r}")

            # historical data is in list format, so we need to return list data as it is
            if data and isinstance(data, dict):
                if data.get(constants.STATUS) == constants.ERROR or data.get(constants.ERROR_CODE):
                    if r.status_code == 403:
                        # TODO: expire session
                        pass

                    # if data.message does not exist, this will create general exception
                    exp = getattr(ex, str(data.get(constants.ERROR_CODE)), ex.GeneralException)
                    raise exp(data[constants.ERROR_MESSAGE], code=r.status_code)
                return data["data"]
            return data
        elif "octet-stream" in r.headers["content-type"]:
            return r.content
        else:
            raise ex.DataException(
                f"Unknown Content-Type ({r.headers['content-type']}) with response: ({r.content[:1000]!r})")

    @staticmethod
    def _user_agent() -> str:
        return (__title__ + "-python/") + __version__

    @staticmethod
    def _parse_user_limits(data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        time_format = "%H:%M:%S %d-%m-%Y"
        records = []  # type: List[Dict[str, Any]]
        for record in data:
            rec = dict(record.items())
            if rec["cash"]:
                rec["cash"] = float(rec["cash"])
            if rec["dayCash"]:
                rec["dayCash"] = float(rec["dayCash"])
            if rec["blockedAmount"]:
                rec["blockedAmount"] = float(rec["blockedAmount"])
            if rec["unClearedCash"]:
                rec["unClearedCash"] = float(rec["unClearedCash"])
            if rec["brokerCollateralAmount"]:
                rec["brokerCollateralAmount"] = float(rec["brokerCollateralAmount"])
            if rec["payIn"]:
                rec["payIn"] = float(rec["payIn"])
            if rec["payOut"]:
                rec["payOut"] = float(rec["payOut"])
            if rec["marginUsed"]:
                rec["marginUsed"] = float(rec["marginUsed"])
            if rec["requestTime"]:
                rec["requestTime"] = dateutil.parser.parse(rec["requestTime"])
            records.append(rec)
        return records

    @staticmethod
    def _parse_greeks(tokens: List[int], greeks: List[Dict]) -> Dict[int, Dict]:
        res = {}  # type: Dict[int, Dict]
        print(tokens, greeks)
        for idx, greek in enumerate(greeks):
            token = tokens[idx]
            res[token] = greek
        return res

    @staticmethod
    def _parse_option_chain(data: List[Dict[str, str]]) -> List[Dict[str, Any]]:
        records = []  # type: List[Dict[str, Any]]
        for row in data:
            rec = dict(row.items())  # type: Dict[str, Any]
            if rec["lotSize"]:
                rec["lotSize"] = int(rec["lotSize"])
            if rec["tickSize"]:
                rec["tickSize"] = float(rec["tickSize"])
            if rec["pricePrecision"]:
                rec["pricePrecision"] = int(rec["pricePrecision"])
            records.append(rec)
        return records

    @staticmethod
    def _parse_historical_data(data: List[Any]) -> List[Dict[str, Any]]:
        records = []
        for row in data:
            if len(row) < 6:
                continue
            record = {
                "time": dateutil.parser.parse(row[0]),  # 2024-03-22T09:15:00+0530
                "open": row[1],
                "high": row[2],
                "low": row[3],
                "close": row[4],
                "volume": row[5]
            }  # type: Dict[str, Any]
            records.append(record)
        return records

    @staticmethod
    def _format_order_by_id_response(data: Union[Dict[str, Any], List[Dict[str, Any]]]) -> List[Dict[str, Any]]:
        _list = []
        if isinstance(data, list):
            _list = data
        elif isinstance(data, dict):
            _list = [data]

        for item in _list:
            # convert date time string to
            for field in ["timeStamp", "exchangeUpdateTime", "requestTime"]:
                if item.get(field) and len(item[field]) == 19:
                    item[field] = dateutil.parser.parse(item[field])

            # convert epoch time to int
            if item.get("orderTime"):
                item["orderTime"] = int(item["orderTime"])
        return _list

    @staticmethod
    def _format_user_orders_response(data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        _list = data
        for item in data:
            # convert date time string to
            for field in ["timeStamp", "exchangeUpdateTime"]:
                if item.get(field) and len(item[field]) == 19:
                    item[field] = dateutil.parser.parse(item[field])

            # convert epoch time to int
            if item.get("orderTime"):
                item["orderTime"] = int(item["orderTime"])
        return data

    @staticmethod
    def _format_user_trades_response(data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        _list = data
        for item in data:
            # convert date time string to
            for field in ["fillTime", "timeStamp", "exchangeUpdateTime", "requestTime"]:
                if item.get(field) and len(item[field]) == 19:
                    item[field] = dateutil.parser.parse(item[field])
        return data

    @property
    def token(self) -> str:
        return self._token or ""

    @token.setter
    def token(self, value: str) -> None:
        raise AttributeError("Cannot set my_var directly. Use set_my_var method instead.")



class TiqsException(Exception):
    def __init__(self, message: str, code: int = 500):
        """Initialize the exception."""
        super(TiqsException, self).__init__(message)
        self.code = code


class GeneralException(TiqsException):
    def __init__(self, message: str, code: int = 500):
        super(GeneralException, self).__init__(message, code)


class DataException(TiqsException):
    def __init__(self, message: str, code: int = 502):
        """Initialize the exception."""
        super(DataException, self).__init__(message, code)



from typing import Optional, Any, Dict, List, Callable, Union

import six
import sys
import time
import json
import logging
import threading

from autobahn.websocket import ConnectionResponse
from twisted.internet import reactor, ssl
from twisted.internet.interfaces import IConnector
from twisted.python import log as twisted_log
from twisted.internet.protocol import ReconnectingClientFactory
from autobahn.twisted.websocket import WebSocketClientProtocol, WebSocketClientFactory, connectWS
from twisted.python.failure import Failure

from .__version__ import __version__, __title__

log = logging.getLogger(__name__)


class TiqsClientProtocol(WebSocketClientProtocol):
    PING_INTERVAL = 2.5
    KEEPALIVE_INTERVAL = 5

    _next_ping = None
    _next_pong_check = None
    _last_pong_time = None
    _last_ping_time = None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(TiqsClientProtocol, self).__init__(*args, **kwargs)

    def onConnect(self, response: ConnectionResponse) -> None:
        self.factory.ws = self
        if self.factory.on_connect:
            self.factory.on_connect(self, response)
        self.factory.resetDelay()
        # pass

    def onOpen(self) -> None:
        self._loop_ping()
        self._loop_pong_check()
        if self.factory.on_open:
            self.factory.on_open(self)

    def onMessage(self, payload: bytes, is_binary: bool) -> None:
        if self.factory.on_message:
            self.factory.on_message(self, payload, is_binary)

    def onClose(self, was_clean: bool, code: int, reason: str) -> None:
        if not was_clean:
            if self.factory.on_error:
                self.factory.on_error(self, code, reason)

        if self.factory.on_close:
            self.factory.on_close(self, code, reason)
        self._last_ping_time = None
        self._last_pong_time = None
        if self._next_ping:
            self._next_ping.cancel()
        if self._next_pong_check:
            self._next_pong_check.cancel()

    def onPong(self, response: str) -> None:
        if self._last_pong_time and self.factory.debug:
            log.debug("last pong was {} seconds back.".format(time.time() - self._last_pong_time))
        self._last_pong_time = time.time()
        if self.factory.debug:
            log.debug("pong => {}".format(response))

    def _loop_ping(self) -> None:
        if self.factory.debug:
            if self._last_ping_time:
                log.debug("last ping was {} seconds back.".format(time.time() - self._last_ping_time))
        self._last_ping_time = time.time()
        self._next_ping = self.factory.reactor.callLater(self.PING_INTERVAL, self._loop_ping)

    def _loop_pong_check(self) -> None:
        if self._last_pong_time:
            last_pong_diff = time.time() - self._last_pong_time
            if last_pong_diff > (2 * self.PING_INTERVAL):
                if self.factory.debug:
                    log.debug("Last pong was {} seconds ago. So dropping connection to reconnect.".format(
                        last_pong_diff))
                self.dropConnection(abort=True)
        self._next_pong_check = self.factory.reactor.callLater(self.PING_INTERVAL, self._loop_pong_check)


class TiqsClientFactory(WebSocketClientFactory, ReconnectingClientFactory):  # type: ignore
    protocol = TiqsClientProtocol
    maxDelay = 5
    maxRetries = 10  # type: Any
    _last_connection_time = None

    def __init__(self, *args: Any, **kwargs: Any):
        self.debug = False
        self.ws = None
        self.on_open = None  # type: Optional[Callable[[websocket.WebSocket], None]]
        self.on_error = None  # type: Optional[Callable[[websocket.WebSocket, int, str], None]]
        self.on_close = None  # type: Optional[Callable[[websocket.WebSocket, int, str], None]]
        self.on_message = None  # type: Optional[Callable[[websocket.WebSocket, bytes, bool], None]]
        self.on_connect = None  # type: Optional[Callable[[websocket.WebSocket, ConnectionResponse], None]]
        self.on_reconnect = None  # type: Optional[Callable[[int], None]]
        self.on_noreconnect = None  # type: Optional[Callable[[], None]]
        super(TiqsClientFactory, self).__init__(*args, **kwargs)

    def startedConnecting(self, connector: IConnector) -> None:
        if not self._last_connection_time and self.debug:
            log.debug("Start WebSocket connection.")
        self._last_connection_time = time.time()

    def clientConnectionFailed(self, connector: IConnector, reason: Failure) -> None:
        if self.retries > 0:
            log.error(
                "Retrying connection. Retry attempt count: {}. Next retry in around: {} seconds".format(self.retries,
                                                                                                        int(round(
                                                                                                            self.delay))))
            if self.on_reconnect:
                self.on_reconnect(self.retries)
        self.retry(connector)  # type: ignore
        self.send_noreconnect()

    def clientConnectionLost(self, connector: IConnector, reason: Failure) -> None:
        if self.retries > 0:
            if self.on_reconnect:
                self.on_reconnect(self.retries)
        self.retry(connector)  # type: ignore
        self.send_noreconnect()

    def send_noreconnect(self) -> None:
        if self.maxRetries is not None and (self.retries > self.maxRetries):
            if self.debug:
                log.debug("Maximum retries ({}) exhausted.".format(self.maxRetries))
                self.stop()
            if self.on_noreconnect:
                self.on_noreconnect()


class TiqsSocket(object):
    CONNECT_TIMEOUT = 30
    RECONNECT_MAX_DELAY = 60
    RECONNECT_MAX_TRIES = 50
    ROOT_URI = "wss://wss.tiqs.trading"

    # Available streaming modes.
    MODE_LTPC = "ltpc"
    MODE_QUOTE = "quote"
    MODE_LTP = "ltp"
    MODE_FULL = "full"

    _is_first_connect = True

    _message_code = 11
    _message_subscribe = "sub"
    _message_unsubscribe = "unsub"
    _message_setmode = "mode"

    _minimum_reconnect_max_delay = 5
    _maximum_reconnect_max_tries = 300

    def __init__(self, app_id: str, token: str, debug: bool = False, root: Optional[str] = None,
                 reconnect: bool = True, reconnect_max_tries: int = RECONNECT_MAX_TRIES,
                 reconnect_max_delay: int = RECONNECT_MAX_DELAY,
                 connect_timeout: int = CONNECT_TIMEOUT) -> None:
        self.root = root or self.ROOT_URI
        if reconnect_max_tries > self._maximum_reconnect_max_tries:
            log.warning(
                "`reconnect_max_tries` can not be more than {val}. Setting to highest possible value - {val}.".format(
                    val=self._maximum_reconnect_max_tries))
            self.reconnect_max_tries = self._maximum_reconnect_max_tries
        else:
            self.reconnect_max_tries = reconnect_max_tries
        if reconnect_max_delay < self._minimum_reconnect_max_delay:
            log.warning(
                "`reconnect_max_delay` can not be less than {val}. Setting to lowest possible value - {val}.".format(
                    val=self._minimum_reconnect_max_delay))
            self.reconnect_max_delay = self._minimum_reconnect_max_delay
        else:
            self.reconnect_max_delay = reconnect_max_delay
        self.connect_timeout = connect_timeout
        self.socket_url = "{root}?appId={app_id}&token={token}".format(
            root=self.root,
            app_id=app_id,
            token=token
        )
        self.debug = debug
        self.ws = None

        self.on_ticks = None  # type: Optional[Callable[[Any, Any], None]]
        self.on_open = None  # type: Optional[Callable[[Any], None]]
        self.on_close = None  # type: Optional[Callable[[Any, int, str], None]]
        self.on_error = None  # type: Optional[Callable[[Any, int, str], None]]
        self.on_connect = None  # type: Optional[Callable[[Any, Any], None]]
        self.on_message = None  # type: Optional[Callable[[Any, bytes, bool], None]]
        self.on_reconnect = None  # type: Optional[Callable[[Any, int], None]]
        self.on_noreconnect = None  # type: Optional[Callable[[Any], None]]
        self.on_order_update = None  # type: Optional[Callable[[Any, Dict[str, Any]], None]]
        self.subscribed_tokens = {}  # type: Dict[int, str]

    def _create_connection(self, url: str, **kwargs: Any) -> None:
        self.factory = TiqsClientFactory(url, **kwargs)
        self.ws = self.factory.ws
        self.factory.debug = self.debug

        # Register private callbacks
        self.factory.on_open = self._on_open
        self.factory.on_error = self._on_error
        self.factory.on_close = self._on_close
        self.factory.on_message = self._on_message
        self.factory.on_connect = self._on_connect
        self.factory.on_reconnect = self._on_reconnect
        self.factory.on_noreconnect = self._on_noreconnect

        self.factory.maxDelay = self.reconnect_max_delay
        self.factory.maxRetries = self.reconnect_max_tries

    def _user_agent(self) -> str:
        return (__title__ + "-python/").capitalize() + __version__

    def connect(self, threaded: bool = False, disable_ssl_verification: bool = False, proxy: Any = None) -> None:
        if self.debug:
            log.debug("Connecting to {}".format(self.socket_url))
        self._create_connection(self.socket_url,
                                useragent=self._user_agent(),
                                proxy=proxy)
        context_factory = None
        if self.factory.isSecure and not disable_ssl_verification:
            context_factory = ssl.ClientContextFactory()
        connectWS(self.factory, contextFactory=context_factory, timeout=self.connect_timeout)
        if self.debug:
            twisted_log.startLogging(sys.stdout)
        opts = {}
        if not reactor.running:
            if threaded:
                opts["installSignalHandlers"] = False
                self.websocket_thread = threading.Thread(target=reactor.run, kwargs=opts)
                self.websocket_thread.daemon = True
                self.websocket_thread.start()
            else:
                reactor.run(**opts)

    def is_connected(self) -> bool:
        if self.ws and self.ws.state and self.ws.state == self.ws.STATE_OPEN:
            return True
        else:
            return False

    def _close(self, code: Optional[int] = None, reason: Optional[str] = None) -> None:
        if self.ws:
            self.ws.sendClose(code, reason)

    def close(self, code: Optional[int] = None, reason: Optional[str] = None) -> None:
        self.stop_retry()
        self._close(code, reason)

    def stop(self) -> None:
        reactor.stop()  # type: ignore

    def stop_retry(self) -> None:
        if self.factory:
            self.factory.stopTrying()

    def subscribe(self, instrument_tokens: List[int]) -> None:
        try:
            self.ws.sendMessage(
                # six.b(json.dumps({"a": self._message_subscribe, "v": instrument_tokens}))
                six.b(json.dumps({
                    "code": self._message_subscribe,
                    "mode": self.MODE_QUOTE,
                    self.MODE_QUOTE: instrument_tokens
                }))
            )

            for token in instrument_tokens:
                self.subscribed_tokens[token] = self.MODE_QUOTE

            return True
        except Exception as e:
            self._close(reason="Error while subscribe: {}".format(str(e)))
            raise

    def unsubscribe(self, instrument_tokens: List[int]) -> None:
        try:
            unsubscribe_dict = {}
            for token in instrument_tokens:
                if token in unsubscribe_dict:
                    unsubscribe_dict[self.subscribed_tokens[token]].append(token)
                else:
                    unsubscribe_dict[self.subscribed_tokens[token]] = [token]

            for mode, tokenList in unsubscribe_dict.items():
                self.ws.sendMessage(
                    six.b(json.dumps({
                        "code": self._message_unsubscribe,
                        "mode": mode,
                        mode: tokenList
                    }))
                )

            for token in instrument_tokens:
                try:
                    del (self.subscribed_tokens[token])
                except KeyError:
                    pass

            return True
        except Exception as e:
            self._close(reason="Error while unsubscribe: {}".format(str(e)))
            raise

    def set_mode(self, mode: str, instrument_tokens: List[int]) -> None:
        try:
            # subscribed_instruments = [int(key) for key in self.subscribed_tokens.keys()]
            # self.unsubscribe(subscribed_instruments)
            req = {
                    "code": self._message_subscribe,
                    "mode": mode,
                    mode: instrument_tokens
                }
            if self.debug:
                log.debug("set_mode payload {}".format(json.dumps(req)))
            self.ws.sendMessage(
                six.b(json.dumps(req))
            )

            # Update modes
            for token in instrument_tokens:
                self.subscribed_tokens[token] = mode

            return True
        except Exception as e:
            self._close(reason="Error while setting mode: {}".format(str(e)))
            raise

    def resubscribe(self) -> None:
        modes = {} # type: Dict[str, List[int]]

        for token in self.subscribed_tokens:
            m = self.subscribed_tokens[token]

            if not modes.get(m):
                modes[m] = []

            modes[m].append(token)

        for mode in modes:
            if self.debug:
                log.debug("Resubscribe and set mode: {} - {}".format(mode, modes[mode]))

            self.subscribe(modes[mode])
            self.set_mode(mode, modes[mode])

    def _on_connect(self, ws, response: ConnectionResponse) -> None:
        self.ws = ws
        if self.on_connect:
            self.on_connect(self, response)

    def _on_close(self, ws, code, reason):
        log.error("Connection closed: {} - {}".format(code, str(reason)))

        if self.on_close:
            self.on_close(self, code, reason)

    def _on_error(self, ws, code: int, reason: str) -> None:
        log.error("Connection error: {} - {}".format(code, str(reason)))

        if self.on_error:
            self.on_error(self, code, reason)

    def _on_message(self, ws, payload: bytes, is_binary: bool) -> None:
        if self.on_message:
            self.on_message(self, payload, is_binary)

        # If the message is binary, parse it and send it to the callback.
        if self.on_ticks and is_binary and len(payload) > 4:
            self.on_ticks(self, self._parse_binary(payload))

        # Parse text messages
        if not is_binary:
            self._parse_message(ws, payload)

    def _on_open(self, ws) -> None:
        if not self._is_first_connect:
            self.resubscribe()
        self._is_first_connect = False

        if self.on_open:
            return self.on_open(self)

    def _on_reconnect(self, attempts_count: int) -> None:
        if self.on_reconnect:
            return self.on_reconnect(self, attempts_count)

    def _on_noreconnect(self) -> None:
        if self.on_noreconnect:
            return self.on_noreconnect(self)

    def _parse_message(self, ws, payload: Union[str, bytes]) -> None:
        if type(payload) == bytes:
            payload = payload.decode("utf-8")
        try:
            data = json.loads(payload)
        except ValueError:
            return
        if self.on_order_update and data.get("type") == "orderUpdate":
            self.on_order_update(self, data)

    def _parse_binary(self, data: bytes) -> Dict[str, Any]:
        tick = {}  # type: Dict[str, Any]
        if self.debug:
            log.debug("Parsing binary, len: {}".format(len(data)))
        if len(data) >= 13:
            tick['token'] = int.from_bytes(data[0:4])
            tick['ltp'] = int.from_bytes(data[4:8])
            tick['mode'] = self.MODE_LTP
        if len(data) == 17:
            tick['close'] = int.from_bytes(data[13:17])
            tick['netChange'] = round(((tick['ltp'] - tick['close']) / tick['close']) * 100 + 1e-9, 2) or 0.00
            if tick['ltp'] > tick['close']:
                tick['changeFlag'] = 43  # ascii code for '+'
            elif tick['ltp'] < tick['close']:
                tick['changeFlag'] = 45  # ascii code for '-'
            else:
                tick['changeFlag'] = 32  # no change
            tick['mode'] = self.MODE_LTPC
        if len(data) >= 69:
            tick['ltq'] = int.from_bytes(data[13:17])
            tick['avgPrice'] = int.from_bytes(data[17:21])
            tick['totalBuyQuantity'] = int.from_bytes(data[21:25])
            tick['totalSellQuantity'] = int.from_bytes(data[25:29])
            tick['open'] = int.from_bytes(data[29:33])
            tick['high'] = int.from_bytes(data[33:37])
            tick['close'] = int.from_bytes(data[37:41])
            tick['low'] = int.from_bytes(data[41:45])
            tick['volume'] = int.from_bytes(data[45:49])
            tick['time'] = int.from_bytes(data[53:57])
            tick['oi'] = int.from_bytes(data[57:61])
            tick['oiDayHigh'] = int.from_bytes(data[61:65])
            tick['oiDayLow'] = int.from_bytes(data[65:69])
            tick['netChange'] = round(((tick['ltp'] - tick['close']) / tick['close']) * 100 + 1e-9, 2) or int(0)
            if tick['ltp'] > tick['close']:
                tick['changeFlag'] = 43  # ascii code for '+'
            elif tick['ltp'] < tick['close']:
                tick['changeFlag'] = 45  # ascii code for '-'
            else:
                tick['changeFlag'] = 32  # no change
            tick['mode'] = self.MODE_QUOTE
        if len(data) >= 197:
            tick['ltt'] = int.from_bytes(data[49:53])
            tick['lowerLimit'] = int.from_bytes(data[69:73])
            tick['upperLimit'] = int.from_bytes(data[73:77])
            bids = []
            asks = []
            for i in range(10):
                quantity = int.from_bytes(data[77 + i * 12:81 + i * 12])
                price = int.from_bytes(data[81 + i * 12:85 + i * 12])
                orders = int.from_bytes(data[85 + i * 12:89 + i * 12])
                if i >= 5:
                    asks.append({'price': price, 'quantity': quantity, 'orders': orders})
                else:
                    bids.append({'price': price, 'quantity': quantity, 'orders': orders})
            tick['bids'] = bids
            tick['asks'] = asks
            tick['mode'] = self.MODE_FULL
        if tick.get('close', 0) == 0:
            tick['netChange'] = 0
        return tick



__title__ = "pytiqs"
__description__ = "The official Python client for the Tiqs trading API"
__url__ = ""
__download_url__ = ""
__version__ = "1.0.15"
__author__ = "Butterfly Broking Pvt. Ltd."
__author_email__ = ""
__license__ = ""



#!/usr/bin/env python

import io
import os
from typing import Dict

from setuptools import setup, find_packages

current_dir = os.path.abspath(os.path.dirname(__file__))
about = {}  # type: Dict[str, str]

with open(os.path.join(current_dir, "pytiqs", "__version__.py"), "r", encoding="utf-8") as f:
    exec(f.read(), about)

with io.open('README.md', 'rt', encoding='utf8') as f:
    readme = f.read()


setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type='text/markdown',
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    download_url=about["__download_url__"],
    license=about["__license__"],
    packages=["pytiqs"],
    install_requires=[
        "service_identity>=18.1.0",
        "requests>=2.18.4",
        "python-dateutil>=2.6.1",
        "six>=1.11.0",
        "pyOpenSSL>=17.5.0",
        "python-dateutil>=2.6.1",
        "autobahn[twisted]==19.11.2"
    ],
    tests_require=["pytest", "responses", "pytest-cov", "mock", "flake8"],
    test_suite="tests",
    setup_requires=["pytest-runner"]
)



from enum import Enum
from typing import Optional

STATUS = "status"  # type: str
ERROR_CODE = "errorCode"  # type: str
ERROR_MESSAGE = "message"  # type: str
ERROR = "error"  # type: str


# class ErrorCode(Enum):
#     BAD_REQUEST_ERROR = "BadRequestError"
#     INTERNAL_SERVER_ERROR = "InternalServerError"
#     NETWORK_ERROR = "NetworkError"
#     NOT_FOUND_ERROR = "NotFoundError"
#     SOMETHING_WENT_WRONG_ERROR = "SomethingWrongError"
#     TOO_MANY_REQUESTS_ERROR = "TooManyRequests"
#     INTERNAL_ERROR = "InternalError"

class BaseEnum(Enum):
    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


class Routes:
    _home_url = "https://app.tiqs.in/app/login"
    _root_url = "https://api.tiqs.trading"

    def __init__(self, root_url: Optional[str] = None) -> None:
        if root_url:
            self._root_url = root_url.strip()

    def get_home_url(self) -> str:
        return self._home_url

    def __getattr__(self, name: str) -> str:
        if name in _PATHS.__members__:
            endpoint_path = _PATHS[name].value
            return "{root_url}{endpoint_path}".format(root_url=self._root_url, endpoint_path=endpoint_path)
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")


class _PATHS(BaseEnum):
    WEB_LOGIN_URL = ""

    AUTHENTICATE_TOKEN = "/auth/app/authenticate-token"
    ALL_INSTRUMENTS = "/all"
    ORDER_MARGIN = "/margin/order"
    BASKET_ORDER_MARGIN = "/margin/basket"
    USER_DETAILS = "/user/details"
    HOLIDAYS_LIST = "/info/holidays"
    INDEX_LIST = "/info/index-list"
    OPTION_CHAIN = "/info/option-chain"
    OPTION_CHAIN_SYMBOLS = "/info/option-chain-symbols"
    GREEKS = "/info/greeks"
    GET_ORDER_BY_ID = "/order/{order_id}"
    GET_USER_ORDERS = "/user/orders"
    GET_USER_TRADES = "/user/trades"
    GET_USER_LIMITS = "/user/limits"
    GET_USER_POSITIONS = "/user/positions"
    PLACE_ORDER = "/order/{variety}"
    UPDATE_ORDER_BY_ID = "/order/regular/{order_id}"
    DELETE_ORDER = "/order/regular/{order_id}"
    GET_HISTORICAL_DATA = "/candle/{exchange}/{token}/{interval}"
    QUOTE_DATA = "/info/quote/{mode}"
    QUOTES_DATA = "/info/quotes/{mode}"

class Variety(BaseEnum):
    REGULAR = "regular"
    COVER = "cover"


class Exchange(BaseEnum):
    NSE = "NSE"
    NFO = "NFO"


class TransactionType(BaseEnum):
    BUY = "B"
    SELL = "S"


class OrderType(BaseEnum):
    MARKET = "MKT"
    LIMIT = "LMT"
    STOP_LOSS_LIMIT = "SL-LMT"
    STOP_LOSS_MARKET = "SL-MKT"


class Retention(BaseEnum):
    DAY = "DAY"
    IOC = "IOC"


class ProductType(BaseEnum):
    MIS = "I"
    CNC = "C"
    NRML = "M"
